#include <unistd.h>

#include <QFileDialog>
#include <QFontDialog>
#include <QFile>
#include <QTextStream>
#include <QIcon>

#include <QDebug>

#include "password.h"
#include "addplaylistitem.h"
#include "addurl.h"
#include "utils.h"
#include "settingsui.h"

SettingsDialog::SettingsDialog(QWidget *parent)
    : QDialog(parent)
{
    ui.setupUi(this);

    settings = Settings::load();

    QIcon::setThemeName(Utils::getIconTheme());
    finalizeUi();
    applySettings();

    rMap[QString("none")] = 0;
    rMap[QString("blacklist")] = 1;
    rMap[QString("whitelist")] = 2;
}

void SettingsDialog::finalizeUi()
{
    // FIXME: don't hardcode /share/ path
    ui.headerLabel->setPixmap(QPixmap("/usr/share/pixmaps/linutop-webkiosk.png"));

    ui.startupChk->setChecked(settings->autostarted());
    ui.passwordBtn->setEnabled(false);
    ui.restrictionDeleteBtn->setEnabled(false);

    QItemSelectionModel *model = ui.restrictionsListView->selectionModel();
    connect(model, SIGNAL(selectionChanged(const QItemSelection &, const QItemSelection &)),
            this, SLOT(on_restrictionsSelection_changed(const QItemSelection &, const QItemSelection &)));

    m_localPlaylistView = new TreeView(this);
    ui.playlistLayout->insertWidget(0, m_localPlaylistView, 1);

    ui.localPlaylistDeleteBtn->setEnabled(false);
    ui.localPlaylistUpBtn->setEnabled(false);
    ui.localPlaylistDownBtn->setEnabled(false);

    // icons for buttons
    ui.restrictionAddBtn->setIcon(QIcon::fromTheme("list-add"));
    ui.restrictionDeleteBtn->setIcon(QIcon::fromTheme("list-remove"));
    ui.restrictionImportBtn->setIcon(QIcon::fromTheme("document-import"));
    ui.restrictionExportBtn->setIcon(QIcon::fromTheme("document-export"));

    ui.localPlaylistAddBtn->setIcon(QIcon::fromTheme("list-add"));
    ui.localPlaylistDeleteBtn->setIcon(QIcon::fromTheme("list-remove"));
    ui.localPlaylistUpBtn->setIcon(QIcon::fromTheme("go-up"));
    ui.localPlaylistDownBtn->setIcon(QIcon::fromTheme("go-down"));
    ui.localPlaylistImportBtn->setIcon(QIcon::fromTheme("document-import"));
    ui.localPlaylistExportBtn->setIcon(QIcon::fromTheme("document-export"));
    ui.localPlaylistCookiesBtn->setIcon(QIcon::fromTheme("emblem-favorite"));

    ui.okBtn->setIcon(QIcon::fromTheme("gtk-ok"));
    ui.cancelBtn->setIcon(QIcon::fromTheme("gtk-cancel"));
    ui.helpBtn->setIcon(QIcon::fromTheme("gtk-help"));
    ui.testBtn->setIcon(QIcon::fromTheme("stock_media-play"));

    // temporary
    ui.localPlaylistImportBtn->setEnabled(false);
    ui.localPlaylistExportBtn->setEnabled(false);
}

void SettingsDialog::setText(QLineEdit *obj, const char *path)
{
    obj->setText(settings->get(path).toString());
}

void SettingsDialog::setChecked(QCheckBox *obj, const char *path, bool value)
{
    obj->setChecked(settings->get(path).toBool() == value);
}

void SettingsDialog::setSpin(QSpinBox *obj, const char *path)
{
    obj->setValue(settings->get(path).toInt());
}

void SettingsDialog::applySettings()
{
    QTabBar *bar = ui.tabWidget->tabBar();

    if (settings->get("main/mode") == "kiosk")
    {
        if (settings->get("main/after_web_session") == "session")
        {
            ui.useKioskRadio->setChecked(true);
            bar->setTabEnabled(1, true);
            bar->setTabEnabled(2, false);
        }
        else
        {
            ui.useBothChk->setChecked(true);
            bar->setTabEnabled(1, false);
            bar->setTabEnabled(2, true);
        }
    }
    else
    {
        ui.useDisplayChk->setChecked(true);
        bar->setTabEnabled(1, false);
        bar->setTabEnabled(2, true);
    }

    setChecked(ui.fullscreenChk, "main/fullscreen", true);
    setChecked(ui.screenSaverChk, "main/no_screensaver", true);

    if (! settings->get("main/admin_passwd").toString().isEmpty())
        ui.passwordChk->setChecked(true);

    QString fontStr(settings->get("webpage/default_font").toString());
    QFont font(fontStr);
    ui.fontBtn->setText(QString("%1 (%2)").arg(font.family()).arg(font.pointSize()));

    // webkiosk
    setChecked(ui.hideToolbarChk, "main/hide_toolbar", true);
    setChecked(ui.browseFSChk, "main/can_access_fs", true);
    setChecked(ui.displayClearChk, "main/can_end_session", true);
    setChecked(ui.sessionAtStartupChk, "main/start_with_session", true);
    setText(ui.homeEdit, "main/default_url");
    setSpin(ui.sessionTimeoutSpin, "main/stop_session_after_seconds");

    setChecked(ui.sessionTimeChk, "main/with_timer", true);
    setSpin(ui.sessionTimeSpin, "main/default_timer");
    setText(ui.postSessionUrlEdit, "main/post_session_uri");
    setSpin(ui.postSessionTimeSpin, "main/post_session_uri_time");

    ui.restrictionsCombo->setCurrentIndex(rMap[settings->get("main/list_mode").toString()]);
    setChecked(ui.regexChk, "main/list_use_regex", true);
    QString f(settings->regexFile());
    restrictionsLoadFromFile(f);

    // display
    setChecked(ui.hideMouseChk, "slideshow/hide_pointer", true);
    setChecked(ui.muteChk, "slideshow/mute", true);
    setSpin(ui.displayTimeSpin, "slideshow/default_time");

    if (settings->get("slideshow/use_remote").toBool())
        ui.updateOnlineRadio->setChecked(true);
    else
        ui.updatePlaylistRadio->setChecked(true);

    setText(ui.remoteEdit, "slideshow/pages");
    setSpin(ui.refreshPlaylistSpin, "slideshow/refresh_after_minutes");
    setChecked(ui.playlistUsbUpdate, "slideshow/do_local_update", true);
    playlistLoadFromFile();
    settings->playlistItems = genPlaylistList();
    connect(m_localPlaylistView->selectionModel(),
            SIGNAL(selectionChanged(QItemSelection, QItemSelection)),
            this, SLOT(on_PlaylistSelectionChanged(QItemSelection, QItemSelection)));

    // network
    switch (settings->get("network/proxy_type").toInt())
    {
        case 1:
            ui.proxyAutoRadio->setChecked(true);
            break;
        case 2:
            ui.proxyManualRadio->setChecked(true);
            break;
        default:
            ui.proxyNoRadio->setChecked(true);
            break;
    }

    setText(ui.proxyUrlEdit, "network/proxy_uri");
    setText(ui.proxyUserEdit, "network/proxy_user");
    setText(ui.proxyPasswordEdit, "network/proxy_passwd");
}

void SettingsDialog::playlistLoadFromFile()
{
    QFile f(settings->playlistFile());
    if (! f.open(QIODevice::ReadOnly | QIODevice::Text))
        return;

    QTextStream in(&f);
    while (!in.atEnd())
    {
        QString line = in.readLine();
        QRegExp rx("^([^#].*)\\s+(\\d+)\\s*$");
        rx.setMinimal(true);
        if (rx.exactMatch(line))
        {
            QStringList list = rx.capturedTexts();
            qDebug() << list;
            m_localPlaylistView->add(list.at(1), list.at(2).toInt());
        }
    }
    f.close();
}

QStringList SettingsDialog::genPlaylistList()
{
    QStringList l;
    QStandardItemModel *m = static_cast<QStandardItemModel *>(m_localPlaylistView->model());

    for (int row=0; row < m->rowCount(); ++row)
    {
        QString uri = m->item(row, 0)->data(Qt::DisplayRole).toString();
        int time = m->item(row, 1)->data(Qt::DisplayRole).toInt();
        l << QString("%1 %2").arg(uri).arg(time);
    }

    return l;
}

bool SettingsDialog::restrictionsLoadFromFile(QString &filename)
{
    ui.restrictionsListView->clear();
    QFile f(filename);
    if (! f.open(QIODevice::ReadOnly | QIODevice::Text))
        return false;

    QTextStream in(&f);
    while (!in.atEnd())
    {
        QString line = in.readLine();
        if (line.isEmpty())
            continue;
        ui.restrictionsListView->addItem(line);
    }
    f.close();

    return true;
}

bool SettingsDialog::restrictionsSaveToFile(QString &filename)
{
    QFile f(filename);
    if (! f.open(QIODevice::WriteOnly | QIODevice::Text))
        return false;

    QTextStream out(&f);
    for (int i=0; i<ui.restrictionsListView->count(); ++i)
        out << ui.restrictionsListView->item(i)->text() << "\n";
    f.close();

    return true;
}

void SettingsDialog::on_restrictionsSelection_changed(
        const QItemSelection &selected,
        const QItemSelection &deselected)
{
    bool b = ui.restrictionsListView->selectionModel()->hasSelection();
    ui.restrictionDeleteBtn->setEnabled(b);
}

void SettingsDialog::on_helpBtn_clicked()
{
    QProcess p(this);
    QStringList args;
    args << "ghelp:linutop-webkiosk";
    p.start("yelp", args);
}

void SettingsDialog::on_testFinished(int exitCode, QProcess::ExitStatus exitStatus)
{
    // rmtree("/tmp/linutop-kiosk-test");
}

void SettingsDialog::on_testBtn_clicked()
{
    // settings->save("/tmp/linutop-kiosk-test");
    QProcess p(this);
    connect(&p, SIGNAL(finished(int, QProcess::ExitStatus)),
            this, SLOT(on_testFinished(int, QProcess::ExitStatus)));
    QStringList args;
    args << "/tmp/linutop-kiosk-test";
    p.start("linutop-webkiosk", args);
}

void SettingsDialog::on_cancelBtn_clicked()
{
    close();
}

void SettingsDialog::on_okBtn_clicked()
{
    if (! ui.passwordChk->isChecked())
        settings->set("main/admin_passwd", QString(""));

    settings->save();
    close();
}

/* main */
void SettingsDialog::on_useBtnGrp_buttonClicked(QAbstractButton *btn)
{
    QTabBar *tb = ui.tabWidget->tabBar();
    if (btn == ui.useKioskRadio)
    {
        tb->setTabEnabled(1, true);
        tb->setTabEnabled(2, false);
        settings->set("main/mode", QString("kiosk"));
        settings->set("main/after_web_session", QString("session"));
        settings->set("main/start_with_session", true);
    }
    else if (btn == ui.useDisplayChk)
    {
        tb->setTabEnabled(1, false);
        tb->setTabEnabled(2, true);
        settings->set("main/mode", QString("display"));
    }
    else
    {
        tb->setTabEnabled(1, true);
        tb->setTabEnabled(2, true);
        settings->set("main/mode", QString("kiosk"));
        settings->set("main/after_web_session", QString("slideshow"));
        if (settings->get("main/with_timer").toBool())
            settings->set("main/start_with_session", false);
    }
}

void SettingsDialog::on_fullscreenChk_toggled(bool checked)
{
    settings->set("main/fullscreen", checked);
}

void SettingsDialog::on_screenSaverChk_toggled(bool checked)
{
    settings->set("main/no_screensaver", checked);
}

void SettingsDialog::on_startupChk_toggled(bool checked)
{
    settings->setAutostarted(checked);
}

void SettingsDialog::on_passwordChk_toggled(bool checked)
{
    ui.passwordBtn->setEnabled(checked);
}

void randomString(char *s, const int len) {
    static const char alphanum[] =
        "0123456789"
        "ABCDEFGHIJKLMNOPQRSTUVWXYZ"
        "abcdefghijklmnopqrstuvwxyz";

    for (int i = 0; i < len; ++i) {
        s[i] = alphanum[rand() % (sizeof(alphanum) - 1)];
    }

    s[len] = 0;
}

void SettingsDialog::on_passwordBtn_clicked()
{
    PasswordDialog d(this);
    if (d.exec() == QDialog::Accepted)
    {
        char salt[10] = "$1$";
        randomString(salt+3, 6);
        char *c = crypt(d.password().toUtf8().data(), salt);
        settings->set("main/admin_passwd", QString(c));
    }
    d.deleteLater();
}

void SettingsDialog::on_fontBtn_clicked()
{
    bool ok;
    QFont font = QFontDialog::getFont(&ok, settings->get("webpage/default_font").toString(), this);
    if (ok)
    {
        settings->set("webpage/default_font", font.toString());
        ui.fontBtn->setText(QString("%1 (%2)").arg(font.family()).arg(font.pointSize()));
    }
}

/* kiosk */
void SettingsDialog::on_hideToolbarChk_toggled(bool checked)
{
    settings->set("main/hide_toolbar", checked);
}

void SettingsDialog::on_browseFSChk_toggled(bool checked)
{
    settings->set("main/can_access_fs", checked);
}

void SettingsDialog::on_displayClearChk_toggled(bool checked)
{
    settings->set("main/can_end_session", checked);
} 

void SettingsDialog::on_sessionAtStartupChk_toggled(bool checked)
{
    settings->set("main/start_with_session", checked);
}

void SettingsDialog::on_homeEdit_textChanged(QString value)
{
    settings->set("main/default_url", value);
}

void SettingsDialog::on_sessionTimeoutSpin_valueChanged(int value)
{
    settings->set("main/stop_session_after_seconds", value);
}

void SettingsDialog::on_sessionTimeChk_toggled(bool checked)
{
    settings->set("main/with_timer", checked);
    settings->set("slideshow/stop_on_event", checked);
}

void SettingsDialog::on_sessionTimeSpin_valueChanged(int value)
{
    settings->set("main/default_timer", value);
}

void SettingsDialog::on_postSessionUrlEdit_textEdited(QString value)
{
    settings->set("main/post_session_uri", value);
}

void SettingsDialog::on_postSessionTimeSpin_valueChanged(int value)
{
    settings->set("main/post_session_uri_time", value);
}

void SettingsDialog::on_restrictionsCombo_currentIndexChanged(int value)
{
    settings->set("main/list_mode", rMap.key(value));
}

void SettingsDialog::on_regexChk_toggled(bool checked)
{
    settings->set("main/list_use_regex", checked);
}

QStringList SettingsDialog::genRestrictionsList()
{
    QStringList l;

    for (int i=0; i < ui.restrictionsListView->count(); ++i)
        l << ui.restrictionsListView->item(i)->data(Qt::DisplayRole).toString();

    qDebug() << l;
    return l;
}

void SettingsDialog::on_restrictionAddBtn_clicked()
{
    AddUrlDialog d;
    int r = d.exec();
    QString url = d.url();
    if ((r == QDialog::Accepted) && (! url.isEmpty()))
            ui.restrictionsListView->addItem(url);

    settings->restrictions = genRestrictionsList();
}

void SettingsDialog::on_restrictionDeleteBtn_clicked()
{
    ui.restrictionsListView->setUpdatesEnabled(false);
    QModelIndex idx = ui.restrictionsListView->selectionModel()->selectedIndexes().at(0);
    ui.restrictionsListView->model()->removeRow(idx.row());
    ui.restrictionsListView->setUpdatesEnabled(true);
    settings->restrictions = genRestrictionsList();
}

void SettingsDialog::on_restrictionImportBtn_clicked()
{
    QString filename = QFileDialog::getOpenFileName(this, "Import");
    if (! filename.isEmpty())
        restrictionsLoadFromFile(filename);
}

void SettingsDialog::on_restrictionExportBtn_clicked()
{
    QString filename = QFileDialog::getOpenFileName(this, "Export");
    if (! filename.isEmpty())
        restrictionsSaveToFile(filename);
}

/* display */
void SettingsDialog::on_hideMouseChk_toggled(bool checked)
{
    settings->set("slideshow/hide_pointer", checked);
}

void SettingsDialog::on_muteChk_toggled(bool checked)
{
    settings->set("slideshow/mute", checked);
}

void SettingsDialog::on_displayTimeSpin_valueChanged(int value)
{
    settings->set("slideshow/default_time", value);
}

void SettingsDialog::on_playlistBtnGrp_buttonClicked(QAbstractButton *btn)
{
    bool b = (btn == ui.updateOnlineRadio);

    settings->set("slideshow/use_remote", b);
    ui.remoteEdit->setEnabled(b);
    ui.refreshPlaylistSpin->setEnabled(b);
    ui.playlistUsbUpdate->setEnabled(!b);
}

void SettingsDialog::on_playlistUsbUpdate_toggled(bool checked)
{
    settings->set("slidehow/do_local_update", checked);
}

void SettingsDialog::on_remoteEdit_textChanged(QString value)
{
    settings->set("slideshow/pages", value);
}

void SettingsDialog::on_refreshPlaylistSpin_valueChanged(int value)
{
    settings->set("slideshow/refresh_after_minutes", value);
}

void SettingsDialog::on_localPlaylistDeleteBtn_clicked()
{
    m_localPlaylistView->delSelection();
    settings->playlistItems = genPlaylistList();
}

void SettingsDialog::on_localPlaylistUpBtn_clicked()
{
    m_localPlaylistView->moveSelection(-1);
    settings->playlistItems = genPlaylistList();
}

void SettingsDialog::on_localPlaylistDownBtn_clicked()
{
    m_localPlaylistView->moveSelection(1);
    settings->playlistItems = genPlaylistList();
}

void SettingsDialog::on_localPlaylistAddBtn_clicked()
{
    AddPlaylistItemDialog d(this);
    if (d.exec() != QDialog::Accepted)
        return;

    QString uri = d.text();
    uri = uri.trimmed();
    if (uri.isEmpty())
        return;

    if (uri.at(0) == '/')
        uri = QString("file://%1").arg(uri);
    int time = d.time();

    m_localPlaylistView->add(uri, time);

    settings->playlistItems = genPlaylistList();
}

void SettingsDialog::on_localPlaylistCookiesBtn_clicked()
{
    // TODO
}

void SettingsDialog::on_PlaylistSelectionChanged(QItemSelection selected, QItemSelection unselected)
{
    bool b = m_localPlaylistView->selectionModel()->hasSelection();
    ui.localPlaylistDeleteBtn->setEnabled(b);
    ui.localPlaylistUpBtn->setEnabled(b);
    ui.localPlaylistDownBtn->setEnabled(b);
}

/* network */
void SettingsDialog::on_proxyBtnGrp_buttonClicked(QAbstractButton *btn)
{
    int t;
    bool b;

    if (btn == ui.proxyAutoRadio)
    {
        t = 1;
        b = false;
    }
    else if (btn == ui.proxyManualRadio)
    {
        t = 2;
        b = true;
    }
    else
    {
        t = 0;
        b = false;
    }

    settings->set("network/proxy_type", t);
    ui.proxyUrlEdit->setEnabled(b);
    ui.proxyUserEdit->setEnabled(b);
    ui.proxyPasswordEdit->setEnabled(b);
}

void SettingsDialog::on_proxyUserEdit_textChanged(QString s)
{
    settings->set("network/proxy_user", s);
}

void SettingsDialog::on_proxyPasswordEdit_textChanged(QString s)
{
    settings->set("network/proxy_passwd", s);
}

void SettingsDialog::on_proxyUrlEdit_textChanged(QString s)
{
    settings->set("network/proxy_uri", s);
}
